#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#define MAXDELAY	5
#define MAXREPEAT	5
#define BUFSIZE		128
#define SEQFILE		"seqno"
#define SEMA		"/tmp/mysem"
#define MAXTRIES	5
#define WAITTIME	1
#define P(sem)		if (!my_lock(sem)) \
			    fprintf(stderr, "my_lock failed\n"), exit(6);
#define V(sem)		if (!my_unlock(sem)) \
			    fprintf(stderr, "my_unlock failed\n"), exit(7);

/* Zugriff ueber die Semaphore sem (Datei!) synchronisieren
 * Rueckgabewert: != 0 bei Erfolg
 */
int my_lock(char *sem) {
    int ret, tries = 0;
	/* Datei exklusiv anlegen => falls sie existiert gibt's einen Fehler;
	 * das Ueberpruefen, ob die Datei existiert, und das Anlegen ist eine
	 * _atomare_ Operation bei open() mit O_EXCL! */
    while ((ret = open(sem, O_RDWR | O_CREAT | O_EXCL, 0664)) < 0 && errno == EEXIST) {
	if (++tries > MAXTRIES)	/* ... das ganze nur MAXTRIES Mal versuchen */
	    return 0;
	sleep(WAITTIME);	/* ... und dazwischen immer ein kleines Schlaefchen */
    }
    return ret >= 0;
}

/* Semaphore wieder freigeben (d.h. Datei loeschen)
 * Rueckgabewert: != 0 bei Erfolg
 */
int my_unlock(char *sem) {
    return unlink(sem) == 0;
}

/* Zaehler lesen, delay Sekunden schlafen und dann erhoehen */
void incr(int delay) {
    int fd, seqno, n;
    char buf[BUFSIZE];

    P(SEMA)	/* kritischen Bereich beginnen; Lock anfordern */
    
    puts("BEGIN kritischer Bereich");
    
    if ((fd = open(SEQFILE, O_RDWR)) < 0)
	perror("open"), exit(2);

	/* Dateiinhalt in den Puffer einlesen */
    if ((n = read(fd, buf, BUFSIZE)) < 0)
	perror("read"), exit(3);
	/* ... und mit Null-Byte terminieren */
    buf[n] = '\0';
    
	/* Integer parsen */
    if (sscanf(buf, "%d", &seqno) != 1)
	fprintf(stderr, "sscanf error\n"), exit(4);

    printf("gelesen: %d\n", seqno);

    seqno++;	/* Zaehler erhoehen */

    sleep(delay);	/* delay Sekunden schlafen */
    
    sprintf(buf, "%03d\n", seqno);	/* neuen Zaehler als String */
    n = strlen(buf);
    
    lseek(fd, 0, SEEK_SET);		/* an den Dateianfang gehen */
    
    if (write(fd, buf, n) != n)		/* ... und den Zaehler schreiben */
	perror("write"), exit(5);

    close(fd);
    
    puts("END kritischer Bereich");
    
    V(SEMA)	/* kritischen Bereich verlassen; Lock wieder freigeben */
}

int main(int argc, char **argv) {
    int i, delay;

    if (argc != 2) {
	fprintf(stderr, "Usage: %s <seed>\n", argv[0]);
	exit(1);
    }

    /* Pseudo-Zufallszahlengenerator initialisieren */
    srandom(atoi(argv[1]));

    /* zufaellig zwischen 0 und MAXDELAY Sekunden schlafen */
    sleep(delay = random() % (MAXDELAY+1));
    
    for (i = 1; i <= MAXREPEAT; i++) {
    	/* gemeinsamen Zaehler erhoehen */
	incr(delay);
	/* zufaellig zwischen 0 und MAXDELAY Sekunden schlafen */
	sleep(delay = random() % (MAXDELAY+1));
    }

    return 0;
}
