#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#define MAXDELAY	5
#define MAXREPEAT	5
#define BUFSIZE		128
#define SEQFILE		"seqno"

/* Zaehler lesen, delay Sekunden schlafen und dann erhoehen */
void incr(int delay) {
    int fd, seqno, n;
    char buf[BUFSIZE];

    if ((fd = open(SEQFILE, O_RDWR)) < 0)
	perror("open"), exit(2);

	/* Dateiinhalt in den Puffer einlesen */
    if ((n = read(fd, buf, BUFSIZE)) < 0)
	perror("read"), exit(3);
	/* ... und mit Null-Byte terminieren */
    buf[n] = '\0';
    
	/* Integer parsen */
    if (sscanf(buf, "%d", &seqno) != 1)
	fprintf(stderr, "sscanf error\n"), exit(4);

    printf("gelesen: %d\n", seqno);

    seqno++;	/* Zaehler erhoehen */

    sleep(delay);	/* delay Sekunden schlafen */
    
    sprintf(buf, "%03d\n", seqno);	/* neuen Zaehler als String */
    n = strlen(buf);
    
    lseek(fd, 0, SEEK_SET);		/* an den Dateianfang gehen */
    
    if (write(fd, buf, n) != n)		/* ... und den Zaehler schreiben */
	perror("write"), exit(5);

    close(fd);
}

int main(int argc, char **argv) {
    int i, delay;

    if (argc != 2) {
	fprintf(stderr, "Usage: %s <seed>\n", argv[0]);
	exit(1);
    }

    /* Pseudo-Zufallszahlengenerator initialisieren */
    srandom(atoi(argv[1]));

    /* zufaellig zwischen 0 und MAXDELAY Sekunden schlafen */
    sleep(delay = random() % (MAXDELAY+1));
    
    for (i = 1; i <= MAXREPEAT; i++) {
    	/* gemeinsamen Zaehler erhoehen */
	incr(delay);
	/* zufaellig zwischen 0 und MAXDELAY Sekunden schlafen */
	sleep(delay = random() % (MAXDELAY+1));
    }

    return 0;
}
